@@ -48,6 +48,7 @@ REPLACERS = [
     'LOCATION',
     'COOKIE',
     'LAST_URL',
+    'URL',
     'HEADERS',
     'RESPONSE',
 ]
@@ -153,6 +154,9 @@ class HTTPTestCase(testtools.TestCase):
                 message[k] = self.replace_template(message[k])
             return message
 
+        if isinstance(message, list):
+            return [self.replace_template(line) for line in message]
+
         for replacer in REPLACERS:
             template = '$%s' % replacer
             method = '_%s_replace' % replacer.lower()
@@ -170,6 +174,10 @@ class HTTPTestCase(testtools.TestCase):
 
         return message
 
+    def load_data_file(self, filename):
+        """Read a file from the current test directory."""
+        return self._load_data_file(filename)
+
     def _assert_response(self):
         """Compare the response with expected data."""
         self._test_status(self.test_data['status'], self.response['status'])
@@ -186,13 +194,27 @@ class HTTPTestCase(testtools.TestCase):
         return value
 
     def _environ_replace(self, message):
-        """Replace an indicator in a message with the environment value."""
+        """Replace an indicator in a message with the environment value.
+
+        If value can be a number, cast it as such. If value is a form of
+        "null", "true", or "false" cast it to None, True, False.
+        """
         value = re.sub(self._replacer_regex('ENVIRON'),
                        self._environ_replacer, message)
-        if value == "False":
+        try:
+            if '.' in value:
+                value = float(value)
+            else:
+                value = int(value)
+            return value
+        except ValueError:
+            pass
+        if value.lower() == "false":
             return False
-        if value == "True":
+        if value.lower() == "true":
             return True
+        if value.lower() == "null":
+            return None
         return value
 
     @staticmethod
@@ -209,11 +231,21 @@ class HTTPTestCase(testtools.TestCase):
 
         With cookie data from set-cookie in the prior request.
         """
-        response_cookies = self.prior.response['set-cookie']
+        return re.sub(self._simple_replacer_regex('COOKIE'),
+                      self._cookie_replacer, message)
+
+    def _cookie_replacer(self, match):
+        """Replace a regex match with the cookie of a previous response."""
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        response_cookies = referred_case.response['set-cookie']
         cookies = http_cookies.SimpleCookie()
         cookies.load(response_cookies)
         cookie_string = cookies.output(attrs=[], header='', sep=',').strip()
-        return message.replace('$COOKIE', cookie_string)
+        return cookie_string
 
     def _headers_replace(self, message):
         """Replace a header indicator in a message with that headers value from
@@ -225,7 +257,12 @@ class HTTPTestCase(testtools.TestCase):
     def _header_replacer(self, match):
         """Replace a regex match with the value of a prior header."""
         header_key = match.group('arg')
-        return self.prior.response[header_key.lower()]
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        return referred_case.response[header_key.lower()]
 
     def _last_url_replace(self, message):
         """Replace $LAST_URL in a message.
@@ -234,16 +271,49 @@ class HTTPTestCase(testtools.TestCase):
         """
         return message.replace('$LAST_URL', self.prior.url)
 
+    def _url_replace(self, message):
+        """Replace $URL in a message.
+
+        With the URL used in a previous request.
+        """
+        return re.sub(self._simple_replacer_regex('URL'),
+                      self._url_replacer, message)
+
+    def _url_replacer(self, match):
+        """Replace a regex match with the value of a previous url."""
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        return referred_case.url
+
     def _location_replace(self, message):
         """Replace $LOCATION in a message.
 
-        With the location header from the prior request.
+        With the location header from a previous request.
         """
-        return message.replace('$LOCATION', self.prior.location)
+        return re.sub(self._simple_replacer_regex('LOCATION'),
+                      self._location_replacer, message)
+
+    def _location_replacer(self, match):
+        """Replace a regex match with the value of a previous location."""
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        return referred_case.location
 
     def _load_data_file(self, filename):
         """Read a file from the current test directory."""
-        path = os.path.join(self.test_directory, os.path.basename(filename))
+        path = os.path.join(self.test_directory, filename)
+        has_dir_traversal = os.path.relpath(
+            path, start=self.test_directory).startswith(os.pardir)
+        if has_dir_traversal:
+            raise ValueError(
+                'Attempted loading of data file outside test directory: %s'
+                % filename)
         with open(path, mode='rb') as data_file:
             return data_file.read()
 
@@ -284,23 +354,54 @@ class HTTPTestCase(testtools.TestCase):
         return urlparse.urlunsplit((parsed_url.scheme, parsed_url.netloc,
                                     parsed_url.path, query_string, ''))
 
+    _history_regex = (
+        r"(?:\$HISTORY\[(?P<quote1>['\"])(?P<case>.+?)(?P=quote1)\]\.)??"
+    )
+
     @staticmethod
     def _replacer_regex(key):
         """Compose a regular expression for test template variables."""
-        return r"\$%s\[(?P<quote>['\"])(?P<arg>.+?)(?P=quote)\]" % key
+        case = HTTPTestCase._history_regex
+        return r"%s\$%s\[(?P<quote>['\"])(?P<arg>.+?)(?P=quote)\]" % (
+            case, key)
+
+    @staticmethod
+    def _simple_replacer_regex(key):
+        """Compose a regular expression for simple variable replacement."""
+        case = HTTPTestCase._history_regex
+        return r"%s\$%s" % (case, key)
 
     def _response_replace(self, message):
-        """Replace a content from the prior request with a value."""
-        replacer_class = self.get_content_handler(
-            self.prior.response.get('content-type'))
-        if replacer_class:
-            replacer_func = replacer_class.gen_replacer(self)
+        """Replace a content path with the value from a previous response.
+
+        If the match would replace the entire message, then don't cast it
+        to a string.
+        """
+        regex = self._replacer_regex('RESPONSE')
+        match = re.match('^%s$' % regex, message)
+        if match:
+            return self._response_replacer(match, preserve=True)
+        return re.sub(regex, self._response_replacer, message)
+
+    def _response_replacer(self, match, preserve=False):
+        """Replace a regex match with the value from a previous response."""
+        response_path = match.group('arg')
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
         else:
-            # If no handler can be found use the null replacer,
-            # which returns "foo" when "$RESPONSE['foo']".
-            replacer_func = base.ContentHandler.gen_replacer(self)
-        return re.sub(self._replacer_regex('RESPONSE'),
-                      replacer_func, message)
+            referred_case = self.prior
+        replacer_class = self.get_content_handler(
+            referred_case.response.get('content-type'))
+        # If no handler can be found use the null replacer,
+        # which returns "foo" when "$RESPONSE['foo']".
+        replacer_class = replacer_class or base.ContentHandler
+        result = replacer_class.replacer(
+            referred_case.response_data, response_path)
+        if preserve:
+            return result
+        else:
+            return six.text_type(result)
 
     def _run_request(self, url, method, headers, body, redirect=False):
         """Run the http request and decode output.
@@ -360,16 +461,22 @@ class HTTPTestCase(testtools.TestCase):
                 raise exception.GabbiFormatError(
                     'malformed headers in test %s: %s' % (test['name'], exc))
 
-        if test['data'] is not '':
+        if test['data'] != '':
             body = self._test_data_to_string(
                 test['data'],
                 utils.extract_content_type(headers, default='')[0])
         else:
             body = ''
 
+        # ensure body is bytes, encoding as UTF-8 because that's
+        # what we do here
+        if isinstance(body, six.text_type):
+            body = body.encode('UTF-8')
+
         if test['poll']:
-            count = test['poll'].get('count', 1)
-            delay = test['poll'].get('delay', 1)
+            count = int(float(self.replace_template(
+                test['poll'].get('count', 1))))
+            delay = float(self.replace_template(test['poll'].get('delay', 1)))
             failure = None
             while count:
                 try:
@@ -402,19 +509,16 @@ class HTTPTestCase(testtools.TestCase):
         """
         if isinstance(data, str):
             if data.startswith('<@'):
-                info = self._load_data_file(data.replace('<@', '', 1))
+                info = self.load_data_file(data.replace('<@', '', 1))
                 if utils.not_binary(content_type):
-                    try:
-                        info = str(info, 'UTF-8')
-                    except TypeError:
-                        info = info.encode('UTF-8')
-                    data = info
+                    data = six.text_type(info, 'UTF-8')
                 else:
                     return info
         else:
             dumper_class = self.get_content_handler(content_type)
             if dumper_class:
-                data = dumper_class.dumps(data)
+                data = self.replace_template(data)
+                data = dumper_class.dumps(data, test=self)
             else:
                 raise ValueError(
                     'unable to process data to %s' % content_type)
@@ -473,7 +577,7 @@ class HTTPTestCase(testtools.TestCase):
                 dumper_class = self.get_content_handler(self.content_type)
                 if dumper_class:
                     full_response = dumper_class.dumps(self.response_data,
-                                                       pretty=True)
+                                                       pretty=True, test=self)
                 else:
                     full_response = self.output
             else:
